@@ -272,8 +272,9 @@ ORBit_marshal_value (GIOPSendBuffer *buf,
 }
 
 static glong
-ORBit_get_union_switch (CORBA_TypeCode tc, gconstpointer *val, 
-		       gboolean update)
+ORBit_get_union_switch (CORBA_TypeCode  tc,
+			gconstpointer  *val, 
+			gboolean        update)
 {
 	glong retval = 0; /* Quiet gcc */
 
@@ -283,23 +284,37 @@ ORBit_get_union_switch (CORBA_TypeCode tc, gconstpointer *val,
 	switch (tc->kind) {
 	case CORBA_tk_ulong:
 	case CORBA_tk_long:
-	case CORBA_tk_enum:
-		retval = *(CORBA_long *)*val;
-		if (update) *val = ((guchar *)*val) + sizeof (CORBA_long);
+	case CORBA_tk_enum: {
+		CORBA_long tmp;
+
+		memcpy (&tmp, *val, sizeof (CORBA_long));
+		retval = (glong) tmp;
+
+		if (update)
+			*val = ((guchar *)*val) + sizeof (CORBA_long);
 		break;
+	}
 	case CORBA_tk_ushort:
-	case CORBA_tk_short:
-		retval = *(CORBA_short *)*val;
-		if (update) *val = ((guchar *)*val) + sizeof (CORBA_short);
+	case CORBA_tk_short: {
+		CORBA_short tmp;
+
+		memcpy (&tmp, *val, sizeof (CORBA_short));
+		retval = (glong) tmp;
+
+		if (update)
+			*val = ((guchar *)*val) + sizeof (CORBA_short);
 		break;
+	}
 	case CORBA_tk_char:
 	case CORBA_tk_boolean:
 	case CORBA_tk_octet:
 		retval = *(CORBA_octet *)*val;
-		if (update) *val = ((guchar *)*val) + sizeof (CORBA_char);
+		if (update)
+			*val = ((guchar *)*val) + sizeof (CORBA_char);
 		break;
 	default:
-		g_error ("Wow, some nut has passed us a weird type[%d] as a union discriminator!", tc->kind);
+		g_error ("Wow, some nut has passed us a weird "
+			 "type[%d] as a union discriminator!", tc->kind);
 	}
 
 	return retval;
@@ -319,9 +334,9 @@ ORBit_get_union_tag (CORBA_TypeCode union_tc,
 		     gconstpointer *val, 
 		     gboolean       update)
 {
-	glong discrim_val, case_val;
-	int i;
 	CORBA_TypeCode retval = CORBA_OBJECT_NIL;
+	glong          discrim_val;
+	int            i;
 
 	discrim_val = ORBit_get_union_switch (
 		union_tc->discriminator, val, update);
@@ -330,11 +345,7 @@ ORBit_get_union_tag (CORBA_TypeCode union_tc,
 		if (i == union_tc->default_index)
 			continue;
 
-		case_val = ORBit_get_union_switch (
-			union_tc->sublabels[i]._type,
-			(gconstpointer*)&union_tc->sublabels[i]._value, FALSE);
-
-		if (case_val == discrim_val) {
+		if (union_tc->sublabels [i] == discrim_val) {
 			retval = union_tc->subtypes[i];
 			break;
 		}
@@ -371,11 +382,10 @@ ORBit_marshal_any (GIOPSendBuffer *buf, const CORBA_any *val)
 /* FIXME: we need two versions of this - one for swap
  * endianness, and 1 for not */
 gboolean
-ORBit_demarshal_value (CORBA_TypeCode tc,
-		      gpointer *val,
-		      GIOPRecvBuffer *buf,
-		      gboolean dup_strings,
-		      CORBA_ORB orb)
+ORBit_demarshal_value (CORBA_TypeCode  tc,
+		       gpointer       *val,
+		       GIOPRecvBuffer *buf,
+		       CORBA_ORB       orb)
 {
 	CORBA_long i;
 
@@ -401,8 +411,7 @@ ORBit_demarshal_value (CORBA_TypeCode tc,
 	}
 	case CORBA_tk_long:
 	case CORBA_tk_ulong:
-	case CORBA_tk_enum:
-	{
+	case CORBA_tk_enum: {
 		CORBA_unsigned_long *ptr;
 		*val = ALIGN_ADDRESS (*val, ORBIT_ALIGNOF_CORBA_LONG);
 		buf->cur = ALIGN_ADDRESS (buf->cur, sizeof (CORBA_long));
@@ -414,11 +423,10 @@ ORBit_demarshal_value (CORBA_TypeCode tc,
 			*ptr = GUINT32_SWAP_LE_BE (*ptr);
 		buf->cur += sizeof (CORBA_long);
 		*val = ((guchar *)*val) + sizeof (CORBA_long);
+		break;
 	}
-	break;
 	case CORBA_tk_longlong:
-	case CORBA_tk_ulonglong:
-	{
+	case CORBA_tk_ulonglong: {
 		CORBA_unsigned_long_long *ptr;
 		*val = ALIGN_ADDRESS (*val, ORBIT_ALIGNOF_CORBA_LONG_LONG);
 		buf->cur = ALIGN_ADDRESS (buf->cur, sizeof (CORBA_long_long));
@@ -430,10 +438,9 @@ ORBit_demarshal_value (CORBA_TypeCode tc,
 			*ptr = GUINT64_SWAP_LE_BE (*ptr);
 		buf->cur += sizeof (CORBA_long_long);
 		*val = ((guchar *)*val) + sizeof (CORBA_long_long);
+		break;
 	}
-	break;
-	case CORBA_tk_longdouble:
-	{
+	case CORBA_tk_longdouble: {
 		CORBA_long_double *ptr;
 		*val = ALIGN_ADDRESS (*val, ORBIT_ALIGNOF_CORBA_LONG_DOUBLE);
 		buf->cur = ALIGN_ADDRESS (buf->cur, sizeof (CORBA_long_double));
@@ -446,8 +453,8 @@ ORBit_demarshal_value (CORBA_TypeCode tc,
 			*ptr = *(CORBA_long_double *)buf->cur;
 		buf->cur += sizeof (CORBA_long_double);
 		*val = ((guchar *)*val) + sizeof (CORBA_long_double);
+		break;
 	}
-	break;
 	case CORBA_tk_float: {
 		CORBA_float *ptr;
 		*val = ALIGN_ADDRESS (*val, ORBIT_ALIGNOF_CORBA_FLOAT);
@@ -499,7 +506,7 @@ ORBit_demarshal_value (CORBA_TypeCode tc,
 		*val = ALIGN_ADDRESS (*val, ORBIT_ALIGNOF_CORBA_ANY);
 		decoded = *val;
 		decoded->_release = CORBA_FALSE;
-		if (ORBit_demarshal_any (buf, decoded, dup_strings, orb))
+		if (ORBit_demarshal_any (buf, decoded, orb))
 			return TRUE;
 		*val = ((guchar *)*val) + sizeof (CORBA_any);
 		break;
@@ -545,7 +552,7 @@ ORBit_demarshal_value (CORBA_TypeCode tc,
 	case CORBA_tk_struct:
 		*val = ALIGN_ADDRESS (*val, tc->c_align);
 		for (i = 0; i < tc->sub_parts; i++) {
-			if (ORBit_demarshal_value (tc->subtypes[i], val, buf, dup_strings, orb))
+			if (ORBit_demarshal_value (tc->subtypes[i], val, buf, orb))
 				return TRUE;
 		}
 		break;
@@ -556,7 +563,7 @@ ORBit_demarshal_value (CORBA_TypeCode tc,
 		int	        sz = 0;
 
 		discrim = *val = ALIGN_ADDRESS (*val, MAX (tc->discriminator->c_align, tc->c_align));
-		if (ORBit_demarshal_value (tc->discriminator, val, buf, dup_strings, orb))
+		if (ORBit_demarshal_value (tc->discriminator, val, buf, orb))
 			return TRUE;
 
 		subtc = ORBit_get_union_tag (tc, (gconstpointer*)&discrim, FALSE);
@@ -564,13 +571,13 @@ ORBit_demarshal_value (CORBA_TypeCode tc,
 			sz = MAX (sz, ORBit_gather_alloc_info (tc->subtypes[i]));
 
 		body = *val = ALIGN_ADDRESS (*val, tc->c_align);
-		if (ORBit_demarshal_value (subtc, &body, buf, dup_strings, orb))
+		if (ORBit_demarshal_value (subtc, &body, buf, orb))
 			return TRUE;
 
 		/* WATCHOUT: end subtc body may not be end of union */
 		*val = ((guchar *)*val) + sz;
+		break;
 	}
-	break;
 	case CORBA_tk_string:
 	case CORBA_tk_wstring:
 		*val = ALIGN_ADDRESS (*val, ORBIT_ALIGNOF_CORBA_POINTER);
@@ -584,12 +591,7 @@ ORBit_demarshal_value (CORBA_TypeCode tc,
 		if ((buf->cur + i) > buf->end
 		    || (buf->cur + i) < buf->cur)
 			return TRUE;
-		if (dup_strings) {
-			*(char **)*val = CORBA_string_dup (buf->cur);
-		} else {
-			*(((ORBitMemHow *) (buf->cur))-1) = ORBIT_MEMHOW_NONE;
-			*(char **)*val = buf->cur;
-		}
+		*(char **)*val = CORBA_string_dup (buf->cur);
 		*val = ((guchar *)*val) + sizeof (CORBA_char *);
 		buf->cur += i;
 		break;
@@ -608,26 +610,42 @@ ORBit_demarshal_value (CORBA_TypeCode tc,
 		else
 			p->_length = *(CORBA_unsigned_long *)buf->cur;
 		buf->cur += sizeof (CORBA_long);
-		if (tc->subtypes[0]->kind == CORBA_tk_octet
-		    || tc->subtypes[0]->kind == CORBA_tk_boolean
-		    || tc->subtypes[0]->kind == CORBA_tk_char) {
+		if (tc->subtypes[0]->kind == CORBA_tk_octet ||
+		    tc->subtypes[0]->kind == CORBA_tk_boolean ||
+		    tc->subtypes[0]->kind == CORBA_tk_char) {
 			/* This special-casing could be taken further to apply to
 			   all atoms... */
-			if ((buf->cur + p->_length) > buf->end
-			    || (buf->cur + p->_length) < buf->cur)
+			if ((buf->cur + p->_length) > buf->end ||
+			    (buf->cur + p->_length) < buf->cur)
 				return TRUE;
 			p->_buffer = ORBit_alloc_simple (p->_length);
 			memcpy (p->_buffer, buf->cur, p->_length);
 			buf->cur = ((guchar *)buf->cur) + p->_length;
 		} else {
+			CORBA_unsigned_long alloc = 1;
+
 			p->_buffer = ORBit_alloc_tcval (tc->subtypes[0],
-							p->_length);
+							MIN (p->_length, alloc));
 			subval = p->_buffer;
 
-			for (i = 0; i < p->_length; i++)
+			for (i = 0; i < p->_length; i++) {
+				if (i == alloc) {
+					size_t delta = (guchar *)subval - (guchar *)p->_buffer;
+
+					p->_buffer = ORBit_realloc_tcval (
+						p->_buffer, tc->subtypes [0],
+						alloc, MIN (alloc * 2, p->_length));
+					alloc = alloc * 2; /* exponential */
+					subval = p->_buffer + delta;
+				}
+
 				if (ORBit_demarshal_value (tc->subtypes[0], &subval,
-							   buf, dup_strings, orb))
+							   buf, orb)) {
+					CORBA_free (p->_buffer);
+					p->_buffer = NULL;
 					return TRUE;
+				}
+			}
 		}
 
 		*val = ((guchar *)*val) + sizeof (CORBA_sequence_CORBA_octet);
@@ -635,7 +653,7 @@ ORBit_demarshal_value (CORBA_TypeCode tc,
 	}
 	case CORBA_tk_array:
 		for (i = 0; i < tc->length; i++)
-			if (ORBit_demarshal_value (tc->subtypes[0], val, buf, dup_strings, orb))
+			if (ORBit_demarshal_value (tc->subtypes[0], val, buf, orb))
 				return TRUE;
 		break;
 	case CORBA_tk_fixed:
@@ -647,18 +665,16 @@ ORBit_demarshal_value (CORBA_TypeCode tc,
 	return FALSE;
 }
 
-
 gpointer
 ORBit_demarshal_arg (GIOPRecvBuffer *buf,
 		     CORBA_TypeCode tc,
-		     gboolean       dup_strings,
 		     CORBA_ORB      orb)
 {
 	gpointer retval, val;
 
 	retval = val = ORBit_alloc_by_tc (tc);
 
-	ORBit_demarshal_value (tc, &val, buf, dup_strings, orb);
+	ORBit_demarshal_value (tc, &val, buf, orb);
 
 	return retval;
 }
@@ -666,7 +682,6 @@ ORBit_demarshal_arg (GIOPRecvBuffer *buf,
 gboolean
 ORBit_demarshal_any (GIOPRecvBuffer *buf,
 		     CORBA_any      *retval,
-		     gboolean        dup_strings,
 		     CORBA_ORB       orb)
 {
 	gpointer val;
@@ -677,8 +692,7 @@ ORBit_demarshal_any (GIOPRecvBuffer *buf,
 		return TRUE;
 
 	val = retval->_value = ORBit_alloc_by_tc (retval->_type);
-	if (ORBit_demarshal_value (retval->_type, &val,
-				   buf, dup_strings, orb))
+	if (ORBit_demarshal_value (retval->_type, &val, buf, orb))
 		return TRUE;
 
 	return FALSE;
@@ -700,7 +714,6 @@ CORBA_any__freekids (gpointer mem, gpointer dat)
 	return t + 1;
 }
 
-
 CORBA_any *
 CORBA_any__alloc (void)
 {
@@ -774,8 +787,7 @@ ORBit_copy_value_core (gconstpointer *val,
 		*val = ((guchar *)*val) + sizeof (CORBA_octet);
 		*newval = ((guchar *)*newval) + sizeof (CORBA_octet);
 		break;
-	case CORBA_tk_any:
-	{
+	case CORBA_tk_any: {
 		const CORBA_any *oldany;
 		CORBA_any *newany;
 		*val = ALIGN_ADDRESS (*val, ORBIT_ALIGNOF_CORBA_ANY);
@@ -783,14 +795,12 @@ ORBit_copy_value_core (gconstpointer *val,
 		oldany = *val;
 		newany = *newval;
 		newany->_type = ORBit_RootObject_duplicate (oldany->_type);
-		/* XXX are we supposed to do above even if oldany->_release
-		   == FALSE? */
 		newany->_value = ORBit_copy_value (oldany->_value, oldany->_type);
 		newany->_release = CORBA_TRUE;
 		*val = ((guchar *)*val) + sizeof (CORBA_any);
 		*newval = ((guchar *)*newval) + sizeof (CORBA_any);
+		break;
 	}
-	break;
 	case CORBA_tk_Principal:
 		*val = ALIGN_ADDRESS (*val,
 				     MAX (MAX (ORBIT_ALIGNOF_CORBA_LONG,
@@ -825,8 +835,7 @@ ORBit_copy_value_core (gconstpointer *val,
 		for (i = 0; i < tc->sub_parts; i++)
 			ORBit_copy_value_core (val, newval, tc->subtypes[i]);
 		break;
-	case CORBA_tk_union:
-	{
+	case CORBA_tk_union: {
 		CORBA_TypeCode utc = ORBit_get_union_tag (tc, (gconstpointer *)val, FALSE);
 		gint	       union_align = tc->c_align;
 		gint	       discrim_align = MAX (tc->discriminator->c_align, tc->c_align);
@@ -893,6 +902,9 @@ ORBit_copy_value (gconstpointer value, CORBA_TypeCode tc)
 {
 	gpointer retval, newval;
 
+	if (!value)
+		return NULL;
+
 	retval = newval = ORBit_alloc_by_tc (tc);
 	ORBit_copy_value_core (&value, &newval, tc);
 
@@ -948,10 +960,22 @@ ORBit_value_equivalent (gpointer *a, gpointer *b,
 		ALIGN_COMPARE (a, b, float, float,   ORBIT_ALIGNOF_CORBA_FLOAT);
 		ALIGN_COMPARE (a, b, double, double, ORBIT_ALIGNOF_CORBA_DOUBLE);
 
-		ALIGN_COMPARE (a, b, boolean, octet, ORBIT_ALIGNOF_CORBA_OCTET);
 		ALIGN_COMPARE (a, b, char,    octet, ORBIT_ALIGNOF_CORBA_OCTET);
 		ALIGN_COMPARE (a, b, octet,   octet, ORBIT_ALIGNOF_CORBA_OCTET);
 
+	case CORBA_tk_boolean: {
+		gboolean ba, bb;
+
+		*a = ALIGN_ADDRESS (*a, ORBIT_ALIGNOF_CORBA_OCTET);
+		*b = ALIGN_ADDRESS (*b, ORBIT_ALIGNOF_CORBA_OCTET);
+		ba = *(CORBA_octet *) *a;
+		bb = *(CORBA_octet *) *b;
+		*a = ((guchar *) *a) + sizeof (CORBA_octet);
+		*b = ((guchar *) *b) + sizeof (CORBA_octet);
+
+		return (ba && bb) || (!ba && !bb);
+	}
+
 	case CORBA_tk_string:
 		*a = ALIGN_ADDRESS (*a, ORBIT_ALIGNOF_CORBA_POINTER);
 		*b = ALIGN_ADDRESS (*b, ORBIT_ALIGNOF_CORBA_POINTER);
@@ -1030,15 +1054,19 @@ ORBit_value_equivalent (gpointer *a, gpointer *b,
 	}
 
 	case CORBA_tk_union: {
-		gint   union_align = tc->c_align;
-		gint   discrim_align = MAX (tc->discriminator->c_align, tc->c_align);
-		size_t union_size = ORBit_gather_alloc_info (tc);
+		gint     union_align = tc->c_align;
+		gint     discrim_align = MAX (tc->discriminator->c_align, tc->c_align);
+		size_t   union_size = ORBit_gather_alloc_info (tc);
+		gpointer a_orig, b_orig;
 
 		CORBA_TypeCode utc_a = ORBit_get_union_tag (
 			tc, (gconstpointer *)a, FALSE);
 		CORBA_TypeCode utc_b = ORBit_get_union_tag (
 			tc, (gconstpointer *)b, FALSE);
 
+		a_orig = *a;
+		b_orig = *b;
+
 		if (!CORBA_TypeCode_equal (utc_a, utc_b, ev))
 			return FALSE;
 
@@ -1054,8 +1082,8 @@ ORBit_value_equivalent (gpointer *a, gpointer *b,
 		if (!ORBit_value_equivalent (a, b, utc_a, ev))
 			return FALSE;
 
-		*a = ((guchar *) *a) + union_size;
-		*b = ((guchar *) *b) + union_size;
+		*a = ((guchar *) a_orig) + union_size;
+		*b = ((guchar *) b_orig) + union_size;
 		return TRUE;
 	}
 
